=== modified file 'bzrlib/tests/test_smart_transport.py'
--- bzrlib/tests/test_smart_transport.py	2006-09-15 12:31:16 +0000
+++ bzrlib/tests/test_smart_transport.py	2006-09-17 23:05:02 +0000
@@ -276,6 +276,87 @@
         self.assertEqual([('_call', ('get', '/foo'))], client._calls)
 
 
+class InstrumentedClient(smart.SmartStreamClient):
+    """A smart client whose writes are stored to a supplied list."""
+
+    def __init__(self, write_output_list):
+        smart.SmartStreamClient.__init__(self, None)
+        self._write_output_list = write_output_list
+
+    def _ensure_connection(self):
+        """We are never strictly connected."""
+
+    def _write_and_flush(self, bytes):
+        self._write_output_list.append(bytes)
+
+
+class InstrumentedServerProtocol(smart.SmartStreamServer):
+    """A smart server which is backed by memory and saves its write requests."""
+
+    def __init__(self, write_output_list):
+        smart.SmartStreamServer.__init__(self, None, None,
+            memory.MemoryTransport())
+        self._write_output_list = write_output_list
+
+    def _write_and_flush(self, bytes):
+        self._write_output_list.append(bytes)
+
+
+class TestSmartProtocol(tests.TestCase):
+    """Tests for the smart protocol.
+
+    Each test case gets a smart_server and smart_client created during setUp().
+
+    It is planned that the client can be called with self.call_client() giving
+    it an expected server response, which will be fed into it when it tries to
+    read. Likewise, self.call_server will call a servers method with a canned
+    serialised client request. Output done by the client or server for these
+    calls will be captured to self.to_server and self.to_client. Each element
+    in the list is a write call from the client or server respectively.
+    """
+
+    def setUp(self):
+        super(TestSmartProtocol, self).setUp()
+        self.to_server = []
+        self.to_client = []
+        self.smart_client = InstrumentedClient(self.to_server)
+        self.smart_server = InstrumentedServerProtocol(self.to_client)
+
+    def assertOffsetSerialisation(self, expected_offsets, expected_serialised,
+        client, server_protocol):
+        """Check that smart (de)serialises offsets as expected.
+        
+        We check both serialisation and deserialisation at the same time
+        to ensure that the round tripping cannot skew: both directions should
+        be as expected.
+        
+        :param expected_offsets: a readv offset list.
+        :param expected_seralised: an expected serial form of the offsets.
+        :param server: a SmartServer instance.
+        """
+        offsets = server_protocol.smart_server._deserialise_offsets(
+            expected_serialised)
+        self.assertEqual(expected_offsets, offsets)
+        serialised = client._serialise_offsets(offsets)
+        self.assertEqual(expected_serialised, serialised)
+
+    def test_server_offset_serialisation(self):
+        """The Smart protocol serialises offsets as a comma and \n string.
+
+        We check a number of boundary cases are as expected: empty, one offset,
+        one with the order of reads not increasing (an out of order read), and
+        one that should coalesce.
+        """
+        self.assertOffsetSerialisation([], '',
+            self.smart_client, self.smart_server)
+        self.assertOffsetSerialisation([(1,2)], '1,2',
+            self.smart_client, self.smart_server)
+        self.assertOffsetSerialisation([(10,40), (0,5)], '10,40\n0,5',
+            self.smart_client, self.smart_server)
+        self.assertOffsetSerialisation([(1,2), (3,4), (100, 200)],
+            '1,2\n3,4\n100,200', self.smart_client, self.smart_server)
+
+
 # TODO: Client feature that does get_bundle and then installs that into a
 # branch; this can be used in place of the regular pull/fetch operation when
 # coming from a smart server.

=== modified file 'bzrlib/transport/smart.py'
--- bzrlib/transport/smart.py	2006-09-15 23:48:16 +0000
+++ bzrlib/transport/smart.py	2006-09-17 23:50:17 +0000
@@ -179,10 +179,9 @@
     return tuple((a.decode('utf-8') for a in req_line[:-1].split('\x01')))
 
 
-def _send_tuple(to_file, args):
-    # XXX: this will be inefficient.  Just ask Robert.
-    to_file.write('\x01'.join((a.encode('utf-8') for a in args)) + '\n')
-    to_file.flush()
+def _encode_tuple(args):
+    """Encode the tuple args to a bytestream."""
+    return '\x01'.join((a.encode('utf-8') for a in args)) + '\n'
 
 
 class SmartProtocolBase(object):
@@ -191,10 +190,8 @@
     def _send_bulk_data(self, body):
         """Send chunked body data"""
         assert isinstance(body, str)
-        self._out.write('%d\n' % len(body))
-        self._out.write(body)
-        self._out.write('done\n')
-        self._out.flush()
+        bytes = ''.join(('%d\n' % len(body), body, 'done\n'))
+        self._write_and_flush(bytes)
 
     # TODO: this only actually accomodates a single block; possibly should support
     # multiple chunks?
@@ -220,6 +217,19 @@
         else:
             self._translate_error(resp)
 
+    def _serialise_offsets(self, offsets):
+        """Serialise a readv offset list."""
+        txt = []
+        for start, length in offsets:
+            txt.append('%d,%d' % (start, length))
+        return '\n'.join(txt)
+
+    def _write_and_flush(self, bytes):
+        """Write bytes to self._out and flush it."""
+        # XXX: this will be inefficient.  Just ask Robert.
+        self._out.write(bytes)
+        self._out.flush()
+
 
 class SmartStreamServer(SmartProtocolBase):
     """Handles smart commands coming over a stream.
@@ -257,11 +267,10 @@
 
     def _send_tuple(self, args):
         """Send response header"""
-        return _send_tuple(self._out, args)
+        return self._write_and_flush(_encode_tuple(args))
 
     def _send_error_and_disconnect(self, exception):
         self._send_tuple(('error', str(exception)))
-        self._out.flush()
         ## self._out.close()
         ## self._in.close()
 
@@ -306,6 +315,11 @@
         self.args = args
         self.body = body
 
+# XXX: TODO: Create a SmartServerRequest which will take the responsibility
+# for delivering the data for a request. This could be done with as the
+# StreamServer, though that would create conflation between request and response
+# which may be undesirable.
+
 
 class SmartServer(object):
     """Protocol logic for smart server.
@@ -314,6 +328,13 @@
     creates responses.
     """
 
+    # IMPORTANT FOR IMPLEMENTORS: It is important that SmartServer not contain
+    # encoding or decoding logic to allow the wire protocol to vary from the
+    # object protocol: we will want to tweak the wire protocol separate from
+    # the object model, and ideally we will be able to do that without having
+    # a SmartServer subclass for each wire protocol, rather just a Protocol
+    # subclass.
+
     # TODO: Better way of representing the body for commands that take it,
     # and allow it to be streamed into the server.
     
@@ -333,6 +354,7 @@
         return SmartServerResponse(('ok',), backing_bytes)
 
     def _deserialise_optional_mode(self, mode):
+        # XXX: FIXME this should be on the protocol object.
         if mode == '':
             return None
         else:
@@ -369,6 +391,30 @@
                 self._recv_body(),
                 self._deserialise_optional_mode(mode))
 
+    def _deserialise_offsets(self, text):
+        # XXX: FIXME this should be on the protocol object.
+        offsets = []
+        for line in text.split('\n'):
+            if not line:
+                continue
+            start, length = line.split(',')
+            offsets.append((int(start), int(length)))
+        return offsets
+
+    def do_put_non_atomic(self, relpath, mode, create_parent, dir_mode):
+        create_parent_dir = (create_parent == 'T')
+        self._backing_transport.put_bytes_non_atomic(relpath,
+                self._recv_body(),
+                mode=self._deserialise_optional_mode(mode),
+                create_parent_dir=create_parent_dir,
+                dir_mode=self._deserialise_optional_mode(dir_mode))
+
+    def do_readv(self, relpath):
+        offsets = self._deserialise_offsets(self._recv_body())
+        backing_bytes = ''.join(bytes for offset, bytes in
+                             self._backing_transport.readv(relpath, offsets))
+        return SmartServerResponse(('readv',), backing_bytes)
+        
     def do_rename(self, rel_from, rel_to):
         self._backing_transport.rename(rel_from, rel_to)
 
@@ -405,6 +451,9 @@
             return SmartServerResponse(('FileExists', e.path))
         except errors.DirectoryNotEmpty, e:
             return SmartServerResponse(('DirectoryNotEmpty', e.path))
+        except errors.ShortReadvError, e:
+            return SmartServerResponse(('ShortReadvError',
+                e.path, str(e.offset), str(e.length), str(e.actual)))
         except UnicodeError, e:
             # If it is a DecodeError, than most likely we are starting
             # with a plain string
@@ -545,6 +594,12 @@
     type: SmartTCPTransport, etc.
     """
 
+    # IMPORTANT FOR IMPLEMENTORS: SmartTransport MUST NOT be given encoding
+    # responsibilities: Put those on SmartClient or similar. This is vital for
+    # the ability to support multiple versions of the smart protocol over time:
+    # SmartTransport is an adapter from the Transport object model to the 
+    # SmartClient model, not an encoder.
+
     def __init__(self, url, clone_from=None, client=None):
         """Constructor.
 
@@ -652,6 +707,32 @@
                                   self._serialise_optional_mode(mode))
         self._translate_error(resp)
 
+    def put_bytes(self, relpath, upload_contents, mode=None):
+        # FIXME: upload_file is probably not safe for non-ascii characters -
+        # should probably just pass all parameters as length-delimited
+        # strings?
+        resp = self._client._call_with_upload(
+            'put',
+            (self._remote_path(relpath), self._serialise_optional_mode(mode)),
+            upload_contents)
+        self._translate_error(resp)
+
+    def put_bytes_non_atomic(self, relpath, bytes, mode=None,
+                             create_parent_dir=False,
+                             dir_mode=None):
+        """See Transport.put_bytes_non_atomic."""
+        # FIXME: no encoding in the transport!
+        create_parent_str = 'F'
+        if create_parent_dir:
+            create_parent_str = 'T'
+
+        resp = self._client._call_with_upload(
+            'put_non_atomic',
+            (self._remote_path(relpath), self._serialise_optional_mode(mode),
+             create_parent_str, self._serialise_optional_mode(dir_mode)),
+            bytes)
+        self._translate_error(resp)
+
     def put_file(self, relpath, upload_file, mode=None):
         # its not ideal to seek back, but currently put_non_atomic_file depends
         # on transports not reading before failing - which is a faulty
@@ -663,15 +744,12 @@
             upload_file.seek(pos)
             raise
 
-    def put_bytes(self, relpath, upload_contents, mode=None):
-        # FIXME: upload_file is probably not safe for non-ascii characters -
-        # should probably just pass all parameters as length-delimited
-        # strings?
-        resp = self._client._call_with_upload(
-            'put',
-            (self._remote_path(relpath), self._serialise_optional_mode(mode)),
-            upload_contents)
-        self._translate_error(resp)
+    def put_file_non_atomic(self, relpath, f, mode=None,
+                            create_parent_dir=False,
+                            dir_mode=None):
+        return self.put_bytes_non_atomic(relpath, f.read(), mode=mode,
+                                         create_parent_dir=create_parent_dir,
+                                         dir_mode=dir_mode)
 
     def append_file(self, relpath, from_file, mode=None):
         return self.append_bytes(relpath, from_file.read(), mode)
@@ -689,6 +767,49 @@
         resp = self._client._call('delete', self._remote_path(relpath))
         self._translate_error(resp)
 
+    def readv(self, relpath, offsets):
+        if not offsets:
+            return
+
+        offsets = list(offsets)
+
+        sorted_offsets = sorted(offsets)
+        # turn the list of offsets into a stack
+        offset_stack = iter(offsets)
+        cur_offset_and_size = offset_stack.next()
+        coalesced = list(self._coalesce_offsets(sorted_offsets,
+                               limit=self._max_readv_combine,
+                               fudge_factor=self._bytes_to_read_before_seek))
+
+
+        resp = self._client._call_with_upload(
+            'readv',
+            (self._remote_path(relpath),),
+            self._client._serialise_offsets((c.start, c.length) for c in coalesced))
+
+        if resp[0] != 'readv':
+            # This should raise an exception
+            self._translate_error(resp)
+            return
+
+        data = self._client._recv_bulk()
+        # Cache the results, but only until they have been fulfilled
+        data_map = {}
+        for c_offset in coalesced:
+            if len(data) < c_offset.length:
+                raise errors.ShortReadvError(relpath, c_offset.start,
+                            c_offset.length, actual=len(data))
+            for suboffset, subsize in c_offset.ranges:
+                key = (c_offset.start+suboffset, subsize)
+                data_map[key] = data[suboffset:suboffset+subsize]
+            data = data[c_offset.length:]
+
+            # Now that we've read some data, see if we can yield anything back
+            while cur_offset_and_size in data_map:
+                this_data = data_map.pop(cur_offset_and_size)
+                yield cur_offset_and_size[0], this_data
+                cur_offset_and_size = offset_stack.next()
+
     def rename(self, rel_from, rel_to):
         self._call('rename', 
                    self._remote_path(rel_from),
@@ -723,6 +844,9 @@
             raise errors.FileExists(resp[1])
         elif what == 'DirectoryNotEmpty':
             raise errors.DirectoryNotEmpty(resp[1])
+        elif what == 'ShortReadvError':
+            raise errors.ShortReadvError(resp[1], int(resp[2]),
+                                         int(resp[3]), int(resp[4]))
         elif what in ('UnicodeEncodeError', 'UnicodeDecodeError'):
             encoding = str(resp[1]) # encoding must always be a string
             val = resp[2]
@@ -809,7 +933,7 @@
 
     def _send_tuple(self, args):
         self._ensure_connection()
-        _send_tuple(self._out, args)
+        return self._write_and_flush(_encode_tuple(args))
 
     def _send_bulk_data(self, body):
         self._ensure_connection()

